/* \file dpipe.cpp
 *
 *  DRISC Pipeline Simulator
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <iostream>
#include <string>
using namespace std;
#include "strtoken.h"
#include "cycle.h"

int step_cpu(int inst, int &nextpc, cycle &step);

#define DO_PIPELINE
int HALT = 0;

int mem[256];
int nwords = 0;

void read_memory(FILE *in);
void show_memory(FILE *out);
int mem_check(int addr);
int set_mem(int addr, int data);
void init_registers();
void show_registers(FILE *out);
void simulate(FILE *out);
void pipeline(FILE *out);

int main(int argc, char *argv[])
{
	FILE *in, *out;
	string filename;
	errno_t nerr;
	if (argc<2) {
		cout << "usage: dpipe hexfile [output file]\n";
		return EXIT_FAILURE;
	}
	filename = argv[1];
	nerr = fopen_s(&in,filename.c_str(),"r");
	if (nerr) {
		cout << "file: " << filename << " not found\n";
		return EXIT_FAILURE;
	}
	if (argc>2) {
		filename = argv[2];
		nerr = fopen_s(&out,filename.c_str(),"wt");
		if (nerr) {
			cout << "file: " << filename << " could not be opened\n";
			return EXIT_FAILURE;
		}
	}
	else out = stdout;
	// single cycle simulation
	read_memory(in);
	fprintf(out,"<html>\n");
	simulate(out);
	fprintf(out,"<h3>Registers</h3>\n");
	show_registers(out);
	fprintf(out,"<h3>Memory</h3>\n");
	show_memory(out);
	// pipeline
#if defined(DO_PIPELINE)
	rewind(in);
	read_memory(in);
	pipeline(out);
	fclose(in);
	fprintf(out,"<h3>Registers</h3>\n");
	show_registers(out);
	fprintf(out,"<h3>Memory</h3>\n");
	show_memory(out);
#endif
	return EXIT_SUCCESS;
}

void show_memory(FILE *out)
{
	int i,flag=0;
	fprintf(out,"<p><table border>\n");
	fprintf(out,"<p><tr bgcolor=CCCCFF><th>addr");
	for (i=0; i<16; i++) fprintf(out,"<th>%X",i);
	fprintf(out,"\n");
	for (i=0; i<nwords; i++) {
		flag = i%16;
		if (flag==0) fprintf(out,"\n<tr align=center><td bgcolor=FFFFCC>%04X",i);
		fprintf(out,"<td>%04X",mem[i]);
		if (flag==7) fprintf(out,"\n");
	}
	while (flag++ < 15) fprintf(out,"<td><br>");
	fprintf(out,"\n</table>\n");
	
}

void read_memory(FILE *in)
{
	char buf[512];
	char *inp, *next;
	int field;
	bool done = false;
	init_registers();
	nwords = 0;
	while (!done) {
		fgets(buf,511,in);
		if (feof(in)) break;
		next = buf;
		while (*next==' ') next++;
		while (next) {
			inp = next;
			while (*inp==' ') inp++; // eat blanks
			next = strqtrm(inp,' ');
			sscanf_s(inp,"%x",&field);
			//if (field>0) printf("%04X\n",field);
			mem[nwords++] = field;
		}
	}
}



int mem_check(int addr)
{
	if (addr<0 || addr>=nwords) {
		fprintf(stderr,"address %x out of range\n",addr);
		return 1;
	}
	return 0;
}

int get_mem(int addr, int &q)
{
	int range_error = mem_check(addr);
	if (range_error) {
		q = 0;
		return 1;
	}
	q = mem[addr];
	return 0;
}

int set_mem(int addr, int data)
{
	int range_error = mem_check(addr);
	if (range_error) return 1;
	mem[addr] = data;
	return 0;
}
	


int regs[] = {0xBAD0,0x1111,0x2222,0x3333,0x4444,0x5555,0x6666,0x7777};

int reg(int n)
{
	if (n>0) return regs[n];
	return 0;
}

void init_registers()
{
	regs[1] = 0x1111;
	regs[2] = 0x2222;
	regs[3] = 0x3333;
	regs[4] = 0x4444;
	regs[5] = 0x5555;
	regs[6] = 0x6666;
	regs[7] = 0x7777;
	regs[8] = 0x8888;
}

void set_reg(int n, int value)
{
	if (n>0) regs[n] = value;
}


void show_registers(FILE *out)
{
	int i;
	fprintf(out,"<p><table border>\n");
	fprintf(out,"<p><tr bgcolor=CCCCFF><th>format");
	for (i=1; i<8; i++) fprintf(out,"<th>r%d",i);
	fprintf(out,"\n");
	fprintf(out,"<tr align=center><th>hex");
	for (i=1; i<8; i++) {
		fprintf(out,"<td>%04X",reg(i));
	}
	fprintf(out,"<tr align=center><th>decimal");
	for (i=1; i<8; i++) {
		fprintf(out,"<td>%d",(short)reg(i));
	}
	fprintf(out,"\n</table>\n");
}



void cycle::set(int pci, int rdi, int qdi)
{
	pc = pci;
	rd = rdi;
	qd = qdi;
	ma = false;
	addr = 0;
}


const int MAX_CYCLES = 256;
cycle steps[MAX_CYCLES];
int nsteps = 0;
	

void simulate(FILE *out)
{
	int nextpc;
	int inst;
	bool done = false;
	if (out) {
		fprintf(out,"<h3>Simulation Results</h3>\n");
		fprintf(out,"<p><table border>\n");
		fprintf(out,"<tr bgcolor=CCCCFF><th>#<th>pc<th>op<th>rd<th>qd<th>addr\n");
	}
	nextpc = 0;
	while (!done) {
		inst = mem[nextpc];
		int ndx = nsteps;
		done = step_cpu(inst,nextpc,steps[ndx]);
		if (out) steps[ndx].show(out,ndx);
		else fprintf(out,"<td><br>\n");
		if (++nsteps>=MAX_CYCLES) break;
		
	}
	if (out) fprintf(out,"</table>\n");
}

void showpc(FILE *out, int pc, bool color)
{
	if (color) fprintf(out,"<td bgcolor=E0E0E0");
	else fprintf(out,"<td");
	fprintf(out," >%04x",pc);
}
		


void parse(FILE *in, FILE *out)
{
	char buf[512];
	char *inp, *next;
	int opcode, ra, rb, rc, field;
	int count, code, n;
	string token;
	bool done = false;
	/*
	 * Go through input file, line by line
	 */
	n = 0;
	while (!done) {
		fgets(buf,511,in);
		if (feof(in)) break;
		//fprintf(out,"%s",buf);
		if (*buf=='#') continue;
		next = buf;
		/*
		 * parse current line
		 */
		count = 0;
		opcode = ra = rb = rc = 0;
		while (next) {
			/*
			 * search for '<', and copy prior text to
			 * output stream
 */
			inp = next;
			next = strqtrm(inp,',');
			sscanf_s(inp,"%d",&field);
			//fprintf(out,"count %d field %d\n",count,field);
			switch (count) {
			case 0:
				token = inp;
				//fprintf(out,"%s %d\n",token.c_str(),opcode);
				count = 1;
				break;
			case 1:
				ra = field;
				count = 2;
				break;
			case 2:
				rb = field;
				count = 3;
				break;
			case 3:
				rc = field;
				count = 4;
				break;
			default:
				break;
			}
		}
		code = ((opcode&0x7)<<13) | ((ra&0x7)<<10) | ((rb&0x7)<<7) | (rc&0x7F);;
		fprintf(out,"%04X\n",code);
		n++;
#if defined(STUFF)
			/*
			 * next points to character after ',' (if ','
			 * was found).
			 */
			/*
			if (next) {
				/*
				*  isolate (in html) text up to the character '>'
				*/
				html = next;
				next = strqtrm(html,'>');
				/*
				*  check for known codes
				*/
				string token = html;
				char *value = strqtrm(html,' ');
				if (token.substr(0,7).compare("include")==0) {
					include(out,value);
					continue; // go to next line
				}
				else if (token.substr(0,4).compare("copy")==0) {
					copy(out,value);
					continue; // go to next line
				}
				else if (token.substr(0,6).compare("insert")==0) {
					parse_insert(value,out);
				}
				/*
				 *  otherwise just echo the html code
				 */
				else fprintf(out,"<%s>",token.c_str());
			}
#endif
	}
	while (n++ < 32) fprintf(out,"0000\n");
}

#if defined(DO_PIPELINE)

IFID IFID_old, IFID_new;
IDEX IDEX_old, IDEX_new;
EXWB EXWB_old, EXWB_new;
WBXX WBXX_old, WBXX_new;

#define FORWARDING

void pipeline(FILE *out)
{
	int nextpc, jump, ncycle=0;
	int ndx = 0;
	int inst, qd;
	bool done = false;
	if (out) {
		fprintf(out,"<h3>Pipeline Results</h3>\n");
		fprintf(out,"<p><table border>\n");
		fprintf(out,"<tr bgcolor=CCCCFF><th>#<th>IF<th>ID<th>EX<th>WB<th>ok\n");
	}
	nextpc = 0;
	ncycle = 0;
	while (!done) {
		// IF stage
		inst =  mem[nextpc];
		IFID_new.active = !IFID_old.halt;
		IFID_new.color = (ncycle&3)==2;
		IFID_new.halt = (inst==HALT) || IFID_old.halt;
		IFID_new.set(inst,nextpc);
		nextpc++;
		// ID stage
		IDEX_new.set(IFID_old);
		//  EX Stage
		EXWB_new.set(IDEX_old);
#if defined(STUFF)
		EXWB_new.active = IDEX_old.active;
		EXWB_new.color = IDEX_old.color;
		EXWB_new.pc = IDEX_old.pc;
		inst = IDEX_old.instr;
		EXWB_new.instr = inst;
		EXWB_new.rd = IDEX_old.rd;
		opcode = inst>>13;
		seim = inst&0x7F;
		im10 = inst&0x3ff;
		if (seim>>6) seim = (0xFFFFFF80)| seim;
		qa = reg(IDEX_old.ra);
		qb = reg(IDEX_old.rb);
#endif
		// data forwarding
#if defined(FORWARDING)
		EXWB_new.forward(IDEX_old,EXWB_old,out,ncycle);
#endif
		//
		jump = EXWB_new.execute(NULL, IFID_new, IDEX_new);
		// WB Stage
		WBXX_new.set(EXWB_old);
#if defined(STUFF)
		WBXX_new.active = EXWB_old.active;
		WBXX_new.color = EXWB_old.color;
		WBXX_new.pc = EXWB_old.pc;
		inst = EXWB_old.instr;
		WBXX_new.instr = inst;
		opcode = inst>>13;
		rd = EXWB_old.rd;
		qd = EXWB_old.qd;
		WBXX_new.rd = rd;
		WBXX_new.qd = qd;
#endif
		//
		if (++ncycle>=MAX_CYCLES) break;
		// show state
		fprintf(out,"<tr align=center><td>%d",ncycle);
		IFID_new.show(out);
		IDEX_new.show(out);
		EXWB_new.show(out);
		WBXX_new.show(out);
		// match to simulation results
		if (WBXX_new.active) {
			qd = steps[ndx].qd;
			int pc = steps[ndx].pc;
			bool flag = WBXX_new.qd==qd &&
				    WBXX_new.pc==pc;
			if (flag) fprintf(out,"<td>");
			else fprintf(out,"<td bgcolor=FF0000>");
			fprintf(out,"%04x<br>%04x\n",pc,qd);
			ndx++;
		}
		else fprintf(out,"<td><br>\n");
		if (EXWB_new.ma) {
			fprintf(out,"<tr align=center><td>%d",++ncycle);
			fprintf(out,"<td colspan=5 align=center>");
			EXWB_new.showmem(out);
		}
		if (EXWB_new.halt) break;
		// change state
		IFID_old = IFID_new;
		IDEX_old = IDEX_new;
		EXWB_old = EXWB_new;
		WBXX_old = WBXX_new;
		if (jump>=0) nextpc = jump;
	}
	if (out) fprintf(out,"</table>\n");
}

#endif
